from .saveable import Saveable
from .animsprite import AnimSprite
from .gamesprite import GameSprite
from .gameinput import InputType, InputDefault, EVENT_TYPE_START_POS
from .modebase import ModeBase
from .fontwrap import FontWrap, get_default_font_wrap
from . import load
from . import utility
from .game import init, stop, get_state, set_state, get_start_mode_cls, get_restart_mode_cls, get_current_mode



import unittest
import os

import pygame

from mode import ModeTest


class TestModeBase(unittest.TestCase):
    DRAW_EXPECTED = "WWWWWb" + os.linesep \
                   + "WrrrrW" + os.linesep \
                   + "WrrgrW" + os.linesep \
                   + "WrrrrW" + os.linesep \
                   + "WrBrrW" + os.linesep \
                   + "WWWWWb" + os.linesep

    @staticmethod
    def get_surface_string(surface: pygame.Surface):
        result = ""
        for y in range(surface.height):
            for x in range(surface.width):
                color = surface.get_at((x, y))
                if color == pygame.Color('white'):
                    result += "W"
                elif color == pygame.Color('red'):
                    result += "r"
                elif color == pygame.Color('green'):
                    result += "g"
                elif color == pygame.Color('blue'):
                    result += "b"
                elif color == pygame.Color('black'):
                    result += "B"
            result += os.linesep
        return result

    @classmethod
    def setUpClass(cls):
        pygame.display.set_mode((1, 1), pygame.NOFRAME)

    def test_draw(self):
        # Arrange
        screen = pygame.Surface((6, 6))
        screen.fill(pygame.Color('white'))
        mode = ModeTest((3, 3))
        mode._camera.topleft = (1, 2)
        # Act
        mode.draw(screen)
        # Assert
        draw_result = self.get_surface_string(screen)
        self.assertEqual(draw_result, self.DRAW_EXPECTED)

    def test_draw_rounding_camera(self):
        # Arrange
        screen = pygame.Surface((6, 6))
        screen.fill(pygame.Color('white'))
        mode = ModeTest((3, 3))
        mode._camera.topleft = (0.9, 1.9)
        # Act
        mode.draw(screen)
        # Assert
        draw_result = self.get_surface_string(screen)
        self.assertEqual(draw_result, self.DRAW_EXPECTED)

    def test_draw_rounding_sprite(self):
        # Arrange
        screen = pygame.Surface((6, 6))
        screen.fill(pygame.Color('white'))
        mode = ModeTest((2.9, 2.9))
        mode._camera.topleft = (1, 2)
        # Act
        mode.draw(screen)
        # Assert
        draw_result = self.get_surface_string(screen)
        self.assertEqual(draw_result, self.DRAW_EXPECTED)

    def test_draw_rounding_both(self):
        # Arrange
        screen = pygame.Surface((6, 6))
        screen.fill(pygame.Color('white'))
        mode = ModeTest((2.9, 2.9))
        mode._camera.topleft = (0.9, 1.9)
        # Act
        mode.draw(screen)
        # Assert
        draw_result = self.get_surface_string(screen)
        self.assertEqual(draw_result, self.DRAW_EXPECTED)

    def test_cleanup(self):
        # Arrange
        mode = ModeTest()
        # Act
        mode.cleanup()
        # Assert
        self.assertEqual(len(mode.sprites_all), 0)


if __name__ == '__main__':
    unittest.main()



import abc
from functools import cache
from typing import final, Self

import pygame

from . import load
from . import game
from .saveable import Saveable
from .modebase import ModeBase
from .inputframe import InputFrame, StateChange


class GameSprite(pygame.sprite.Sprite, Saveable, abc.ABC):
    """Base class for many game objects.
    Subclasses should set:
    optional: _IMAGE_LOCATION, location of image file (if not set subclass must set image and rect)
    optional: _ALPHA_OR_COLORKEY, used for loading image, required if _IMAGE_LOCATION is set
    optional: _IMAGE_SECTION_SIZE, used if only displaying subset of image for sprite animation
    optional: _COLLISION_RADIUS, set this if a circle collision is appropriate for this sprite
    optional: _COLLISION_MASK_LOCATION, location of image file for generating a collision mask
    optional: _COLLISION_MASK_ALPHA_OR_COLORKEY, used for loading image, required if _COLLISION_MASK_LOCATION is set
    optional: _GETS_INPUT, set this true to force this sprite to receive input

    To hook in to collision checking, create a function like so:
    def collide_OtherGameSpriteClassName(self, other: OtherGameSpriteClassName):
        do_something()
    collide_OtherGameSpriteClassName will be called whenever there is a collision with a OtherGameSpriteClassName
    Other will be the GameSprite collided with.
    """
    _IMAGE_LOCATION: str = None
    _ALPHA_OR_COLORKEY: bool | tuple[int, int, int] = None
    _IMAGE_SECTION_SIZE: tuple[int, int] | None = None
    _COLLISION_RADIUS: float | None = None
    _COLLISION_MASK_LOCATION: str | None = None
    _COLLISION_MASK_ALPHA_OR_COLORKEY: bool | tuple[int, int, int] | None = None
    _GETS_INPUT: bool = False

    __slots__ = (
        '_input_frame',
        '_base_image',
        '_source_rect',
        '_image_count_x',
        '_image_count_y',
        '_seq',
        '_mask_image',
        '_mask_source_rect',
        '_mask_image_count_x',
        '_mask_image_count_y',
        '_mask_seq',
    )

    def __init__(self, **kwargs):
        if self._IMAGE_LOCATION and not self._ALPHA_OR_COLORKEY:
            raise RuntimeError(
                "if _IMAGE_LOCATION is set, _ALPHA_OR_COLORKEY must be set"
            )
        if self._COLLISION_MASK_LOCATION and not self._COLLISION_MASK_ALPHA_OR_COLORKEY:
            raise RuntimeError(
                "if _COLLISION_MASK_LOCATION is set, _COLLISION_MASK_ALPHA_OR_COLORKEY must be set"
            )
        super().__init__()
        self.image = None
        self.rect = None
        self._input_frame: InputFrame | None = None
        self._base_image: pygame.Surface | None = None
        self._image_count_x: int | None = None
        self._image_count_y: int | None = None
        self._seq: int | None = None
        self._source_rect: pygame.Rect | None = None
        if self._IMAGE_LOCATION:
            self._base_image = load.image(self._IMAGE_LOCATION, self._ALPHA_OR_COLORKEY)
            self.image = self._base_image
            if self._IMAGE_SECTION_SIZE:
                image_size = self._base_image.get_size()
                self._image_count_x = image_size[0] // self._IMAGE_SECTION_SIZE[0]
                self._image_count_y = image_size[1] // self._IMAGE_SECTION_SIZE[1]
                self._seq = 0
                self._source_rect = pygame.Rect((0, 0), self._IMAGE_SECTION_SIZE)
                self.image = load.subsurface(self._base_image, tuple(self._source_rect))
            self.rect = self.image.get_frect(**kwargs)
        self.radius: float | None = None
        self._mask_image: pygame.Surface | None = None
        self._mask_image_count_x: int | None = None
        self._mask_image_count_y: int | None = None
        self._mask_seq: int | None = None
        self._mask_source_rect: pygame.Rect | None = None
        if self.image:
            size = self.image.get_size()
            self.mask = load.mask_filled(size)
            if self._COLLISION_RADIUS:
                self.radius = self._COLLISION_RADIUS
                self.mask = load.mask_circle(size, self.radius)
            if self._COLLISION_MASK_LOCATION:
                self._mask_image = load.image(self._COLLISION_MASK_LOCATION, self._COLLISION_MASK_ALPHA_OR_COLORKEY)
                mask_image_size = self._mask_image.get_size()
                if size != mask_image_size:
                    self._mask_image_count_x = mask_image_size[0] // size[0]
                    self._mask_image_count_y = mask_image_size[1] // size[1]
                    self._mask_seq = 0
                    self._mask_source_rect = pygame.Rect((0, 0), size)
                self.mask = load.mask_surface(self._mask_image, self._mask_source_rect and tuple(self._mask_source_rect))

    def save(self):
        return {
            'rect_topleft': self.rect.topleft,
            '_seq': self._seq,
            '_mask_seq': self._mask_seq,
        }

    @classmethod
    def load(cls, save_data):
        new_obj = GameSprite(topleft=save_data['rect_topleft'])
        if new_obj.seq is not None:
            new_obj.seq = save_data['_seq']
        if new_obj.mask_seq is not None:
            new_obj.mask_seq = save_data['_mask_seq']
        return new_obj

    @final
    @property
    def seq(self):
        """Get the image sequence.
        Set this value to change to a different part of the image sheet."""
        return self._seq

    @final
    @seq.setter
    def seq(self, value: int):
        if self._seq is None:
            raise RuntimeError("error: setting seq for GameSprite not using a sprite sheet")
        self._seq = value % (self._image_count_x * self._image_count_y)
        self._source_rect.x = (self._seq % self._image_count_x) * self._IMAGE_SECTION_SIZE[0]
        self._source_rect.y = (self._seq // self._image_count_x) * self._IMAGE_SECTION_SIZE[1]
        self.image = load.subsurface(self._base_image, tuple(self._source_rect))

    @final
    @property
    def mask_seq(self):
        """Get the mask sequence.
        Set this value to change to a different part of the mask sheet."""
        return self._mask_seq

    @final
    @mask_seq.setter
    def mask_seq(self, value: int):
        if self._mask_seq is None:
            raise RuntimeError("error: setting mask_seq for GameSprite not using a sprite sheet")
        size = self.image.get_size()
        self._mask_seq = value % (self._mask_image_count_x * self._mask_image_count_y)
        self._mask_source_rect.x = (self._mask_seq % self._mask_image_count_x) * size[0]
        self._mask_source_rect.y = (self._mask_seq // self._mask_image_count_x) * size[1]
        self.mask = load.mask_surface(self._mask_image, tuple(self._mask_source_rect))

    @classmethod
    @final
    @cache
    def get_collision_labels(cls):
        labels = [
            t.__name__
            for t in cls.mro()
            if t not in GameSprite.mro()
        ]
        labels.append('GameSprite')
        return frozenset(labels)

    @classmethod
    @final
    @cache
    def get_collides_with(cls):
        return frozenset([
            e.removeprefix('collide_')
            for e in dir(cls)
            if e.startswith('collide_')
        ])

    @final
    def does_collide(self, other: Self):
        if self.radius and other.radius:
            # no rounding since circle collisions doesn't match exact pixels anyway
            dx = self.rect.centerx - other.rect.centerx
            dy = self.rect.centery - other.rect.centery
            ds = dx**2 + dy**2
            return ds <= (self.radius + other.radius)**2
        elif self.radius or other.radius or self._COLLISION_MASK_LOCATION or other._COLLISION_MASK_LOCATION:
            # rounding so that mask collisions reflect apparent (drawn) position of sprites
            dx = round(other.rect.x) - round(self.rect.x)
            dy = round(other.rect.y) - round(self.rect.y)
            return self.mask.overlap(other.mask, (dx, dy))
        else:
            # rounding so that rect collisions reflect apparent (drawn) positions of sprites
            self_rect = self.rect.move_to(topleft=(round(self.rect.x), round(self.rect.y)))
            other_rect = other.rect.move_to(topleft=(round(other.rect.x), round(other.rect.y)))
            return self_rect.colliderect(other_rect)

    @final
    def start(self, mode: ModeBase | None = None):
        """Function to start processing the GameSprite as part of running the game.
        Should usually be called after creating the GameSprite.
        Adds the sprite to appropriate groups in the current game mode or the mode indicated.
        Sprites created within a mode's __init__ should probably have that mode passed into this method.
        """
        if mode is None:
            mode = game.get_current_mode()
        mode.sprites_all.add(self)
        mode.sprites_game.add(self)
        if self._GETS_INPUT or self._take_state_change is not GameSprite._take_state_change:
            mode.sprites_input.add(self)
        self._start(mode)
        return self

    @final
    def input(self, input_frame: InputFrame):
        """Called to pass the current InputFrame to a GameSprite."""
        for state_change in input_frame.state_changes:
            self._take_state_change(state_change)
        self._input_frame = input_frame

    def update(self, dt: int, camera: pygame.FRect):
        """Called to apply time updates to a GameSprite."""
        pass

    def draw_dynamic(self, screen: pygame.Surface, offset: pygame.typing.IntPoint):
        """Called to apply dynamic drawing from a GameSprite after normal drawing."""
        pass

    def _start(self, mode: ModeBase):
        """Called when a GameSprite is started."""
        pass

    def _take_state_change(self, state_change: StateChange):
        """Handle input state change (this is called on all state changes if this GameSprite receives input)
        During this method call self._input_frame still holds the old input_frame.
        Overriding this method ensures that the child class will receive input."""
        pass



import os
from types import ModuleType
from collections.abc import Iterable

import pygame
pygame.init()

from . import display
from . import gameinput
from . import fontwrap
from .modebase import ModeBase
from .modegamemenu import ModeGameMenu
from .modegamemenu import ModeGameMenuTop
from .saveable import Saveable
from . import config
from . import save


class _Game(object):
    __slots__ = (
        '_running',
        'start_mode_cls',
        'restart_mode_cls',
        '_state_cls',
        '_clock',
        '_max_dt',
        '_joysticks',
        '_is_first_loop',
        'current_mode',
        'state',
    )

    def __init__(
        self,
        mode_module: ModuleType,
        start_mode_cls: type[ModeBase],
        restart_mode_cls: type[ModeBase] | None,
        state_cls: type[Saveable],
        src_directory: str,
        screen_size: tuple[int, int],
        title: str,
        window_icon: str | None,
        max_players: int,
        event_names: tuple[str],
        input_defaults: tuple[gameinput.InputDefault],
        font_location: str | None,
        font_size: int,
        font_height: int,
        font_antialias: bool,
        max_dt: int
    ):
        self._running = False
        self.start_mode_cls = start_mode_cls
        self.restart_mode_cls = restart_mode_cls
        self._state_cls = state_cls
        config.init(
            os.path.join(src_directory, 'config.ini')
        )
        save.init(
            os.path.join(src_directory, 'saves'),
            mode_module
        )
        display.init(
            os.path.join(src_directory, 'screenshots'),
            screen_size,
            title,
            window_icon
        )
        gameinput.init(
            os.path.join(src_directory, 'input.cfg'),
            max_players,
            event_names,
            input_defaults
        )
        if font_location:
            font = pygame.font.Font(font_location, font_size)
        else:
            font = pygame.font.SysFont(None, font_size)
        fontwrap.init(font, font_height, font_antialias)
        self._clock = pygame.time.Clock()
        self._max_dt = max_dt
        self._joysticks = [
            pygame.joystick.Joystick(i)
            for i
            in range(pygame.joystick.get_count())
        ]
        self._is_first_loop = True
        self.current_mode: ModeBase | None = None
        self.state: Saveable | None = None

    def start(self):
        """Start the game, must be called before run()."""
        self.set_state()
        self.current_mode = self.start_mode_cls()
        self._running = True

    def set_state(self, save_data=None):
        if save_data:
            self.state = self._state_cls.load(save_data)
        else:
            self.state = self._state_cls()

    def stop(self):
        self._running = False

    def run(self):
        """Run the game, and check if the game needs to end."""
        if not self.current_mode:
            raise RuntimeError("error: self._current_mode is not set")
        events = self._filter_input(pygame.event.get())
        events = gameinput.take_events(events)
        input_frame = gameinput.get_input_frame()
        if input_frame.was_player_input_pressed(0, gameinput.TYPE_SCREENSHOT):
            display.take_screenshot()
        if any(map(self._is_pause_event, events)) or input_frame.was_input_pressed(gameinput.TYPE_PAUSE):
            # if already in pause menu no need to do this stuff
            if not isinstance(self.current_mode, ModeGameMenu):
                self.current_mode = ModeGameMenuTop(self.current_mode)
                gameinput.start_new_mode()
                input_frame = gameinput.get_input_frame()
                pygame.mixer.music.pause()
                pygame.mixer.pause()
                events = []
        self.current_mode.input(events, input_frame)
        dt = self._clock.tick_busy_loop(display.max_framerate)
        while dt > self._max_dt:
            dt -= self._max_dt
            self.current_mode.update(self._max_dt)
        self.current_mode.update(dt)
        self.current_mode.draw(display.screen)
        display.scale_draw()
        if self.current_mode.next_mode is not None:
            if isinstance(self.current_mode, ModeGameMenu) \
                    and not isinstance(self.current_mode.next_mode, ModeGameMenu):
                pygame.mixer.music.unpause()
                pygame.mixer.unpause()
            self.current_mode.cleanup()
            self.current_mode = self.current_mode.next_mode
            gameinput.start_new_mode()
        self._is_first_loop = False
        if not self._running:
            config.save()
            gameinput.save()
            self.current_mode = None
            self.state = None
            pygame.quit()
        return self._running

    def _filter_input(self, events: Iterable[pygame.event.Event]):
        """Take care of input that game modes should not take care of."""
        events = map(display.scale_mouse_input, events)
        events = filter(self._filter_event, events)
        return list(events)

    def _filter_event(self, event: pygame.event.Event):
        """If event should be handled before all others, handle it and return False, otherwise return True.
        As an example, game-ending or display-changing events should be handled before all others.
        Also filter out irrelevant mouse events here.
        """
        match event.type:
            case pygame.MOUSEMOTION | pygame.MOUSEBUTTONUP | pygame.MOUSEBUTTONDOWN:
                return display.is_in_screen(event.pos)
            case pygame.JOYDEVICEREMOVED:
                self._joysticks = [
                    joystick
                    for joystick
                    in self._joysticks
                    if joystick.get_instance_id() != event.instance_id
                ]
                return False
            case pygame.JOYDEVICEADDED:
                self._joysticks = [
                    pygame.joystick.Joystick(i)
                    for i
                    in range(pygame.joystick.get_count())
                ]
                return False
        return True

    def _is_pause_event(self, event: pygame.event.Event):
        return event.type in {pygame.QUIT, pygame.WINDOWFOCUSLOST, pygame.WINDOWMINIMIZED} \
            or (event.type == pygame.WINDOWMOVED and not self._is_first_loop)


_game: _Game | None = None


def init(
    mode_module: ModuleType,
    start_mode_cls: type[ModeBase],
    state_cls: type[Saveable],
    src_directory: str,
    screen_size: tuple[int, int],
    title: str,
    window_icon: str | None,
    max_players: int,
    event_names: tuple[str],
    input_defaults: tuple[gameinput.InputDefault],
    font_location: str | None,
    font_size: int,
    font_height: int,
    font_antialias: bool,
    restart_mode_cls: type[ModeBase] | None=None,
    max_dt: int=5
):
    """Loads up the game and prepares it for running.
    Arguments:
    mode_module: the module holding all modes for your game
    start_mode_cls: the class for the first mode
    state_cls: the class for holding general game state
    src_directory: directory of the program
    screen_size: size of the virtual screen
    title: title of the game (for titlebar)
    window_icon: location of icon of the game (for titlebar)
    max_players: maximum number of players the game supports
    event_names: names for virtual inputs that button / axis inputs map to
    input_defaults: default input mappings
    font_location: location of default font for the game
    font_size: default font size
    font_height: default font height
    font_antialias: default font antialias
    restart_mode_cls: the class to return for what mode to restart the game to
    max_dt: maximum dt for updates, if over this will run updates and collision checks multiple times
    """
    global _game
    if _game:
        raise RuntimeError("error: _game is already set")
    _game = _Game(
        mode_module,
        start_mode_cls,
        restart_mode_cls,
        state_cls,
        src_directory,
        screen_size,
        title,
        window_icon,
        max_players,
        event_names,
        input_defaults,
        font_location,
        font_size,
        font_height,
        font_antialias,
        max_dt
    )
    _game.start()
    return _game


def stop():
    _game.stop()


def get_state():
    return _game.state


def set_state(save_data=None):
    _game.set_state(save_data)


def get_start_mode_cls():
    return _game.start_mode_cls


def get_restart_mode_cls():
    return _game.restart_mode_cls or _game.start_mode_cls


def get_current_mode():
    return _game.current_mode



import unittest

import pygame

from jovialengine.gamesprite import GameSprite


class TestSpriteA(GameSprite):
    pass

class TestSpriteB(TestSpriteA):
    pass

class TestSpriteC(GameSprite):
    _GETS_INPUT = True
    def collide_TestSpriteA(self, other):
        pass

class TestSpriteD(GameSprite):
    def _take_state_change(self, state_change):
        print(state_change)

class TestSpriteCircle(GameSprite):
    _IMAGE_LOCATION = './assets/gfx/4x4_image.png'
    _ALPHA_OR_COLORKEY = (255, 0, 255)
    _COLLISION_RADIUS = 1.5

class TestSpriteRect(GameSprite):
    _IMAGE_LOCATION = './assets/gfx/4x4_image.png'
    _ALPHA_OR_COLORKEY = (255, 0, 255)

class TestSpriteMask(GameSprite):
    _IMAGE_LOCATION = './assets/gfx/4x4_image.png'
    _ALPHA_OR_COLORKEY = (255, 0, 255)
    _COLLISION_MASK_LOCATION = './assets/gfx/4x4_image.png'
    _COLLISION_MASK_ALPHA_OR_COLORKEY = (255, 0, 255)

class TestSpriteSheet(GameSprite):
    _IMAGE_LOCATION = './assets/gfx/6x4_sheet_tests.png'
    _ALPHA_OR_COLORKEY = (255, 0, 255)
    _IMAGE_SECTION_SIZE = (2, 2)
    _COLLISION_MASK_LOCATION = './assets/gfx/6x4_mask_tests.png'
    _COLLISION_MASK_ALPHA_OR_COLORKEY = (255, 0, 255)

class TestGameSprite(unittest.TestCase):
    IMAGE_POINTS = (
        (0, 0),
        (1, 0),
        (0, 1),
        (1, 1),
    )

    @classmethod
    def setUpClass(cls):
        pygame.display.set_mode((1, 1), pygame.NOFRAME)

    def test_get_collision_labels_GameSprite(self):
        # Assert
        self.assertEqual(GameSprite.get_collision_labels(), frozenset(('GameSprite',)))

    def test_get_collision_labels_TestSpriteA(self):
        # Assert
        self.assertEqual(TestSpriteA.get_collision_labels(), frozenset(('TestSpriteA','GameSprite',)))

    def test_get_collision_labels_TestSpriteB(self):
        # Assert
        self.assertEqual(TestSpriteB.get_collision_labels(), frozenset(('TestSpriteB','TestSpriteA','GameSprite',)))

    def test_get_collision_labels_TestSpriteC(self):
        # Assert
        self.assertEqual(TestSpriteC.get_collision_labels(), frozenset(('TestSpriteC','GameSprite',)))

    def test_get_collision_labels_TestSpriteD(self):
        # Assert
        self.assertEqual(TestSpriteD.get_collision_labels(), frozenset(('TestSpriteD','GameSprite',)))

    def test_get_collides_with_TestSpriteA(self):
        # Assert
        self.assertEqual(TestSpriteA.get_collides_with(), frozenset(()))

    def test_get_collides_with_TestSpriteC(self):
        # Assert
        self.assertEqual(TestSpriteC.get_collides_with(), frozenset(('TestSpriteA',)))

    def test_does_collide_circle_false(self):
        # Arrange
        left = TestSpriteCircle(center=(2.5, 3))
        right = TestSpriteCircle(center=(8.5, 4))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_circle_true(self):
        # Arrange
        left = TestSpriteCircle(center=(2.5, 3))
        right = TestSpriteCircle(center=(3.99, 3))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_rect_false(self):
        # Arrange
        left = TestSpriteRect(center=(2.5, 3))
        right = TestSpriteRect(center=(8.5, 4))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_rect_true(self):
        # Arrange
        left = TestSpriteRect(center=(2, 3))
        right = TestSpriteRect(center=(5, 6))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_rect_round_false_0(self):
        # Arrange
        left = TestSpriteRect(topleft=(0.4, 0.0))
        right = TestSpriteRect(topleft=(4.0, 0.0))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_rect_round_false_1(self):
        # Arrange
        left = TestSpriteRect(topleft=(0.4, 0.0))
        right = TestSpriteRect(topleft=(4.0, 0.0))
        # Act
        does_collide = right.does_collide(left)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_rect_round_true_0(self):
        # Arrange
        left = TestSpriteRect(topleft=(0.51, 0.0))
        right = TestSpriteRect(topleft=(4.0, 0.0))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_rect_round_true_1(self):
        # Arrange
        left = TestSpriteRect(topleft=(0.51, 0.0))
        right = TestSpriteRect(topleft=(4.0, 0.0))
        # Act
        does_collide = right.does_collide(left)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_mask_false_0(self):
        # Arrange
        left = TestSpriteMask(center=(2, 3))
        right = TestSpriteMask(center=(5, 3))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_mask_false_1(self):
        # Arrange
        left = TestSpriteMask(center=(3, 3))
        right = TestSpriteMask(center=(5, 5))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_mask_true(self):
        # Arrange
        left = TestSpriteMask(center=(4, 4))
        right = TestSpriteMask(center=(6, 2))
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_mask_round_false_0(self):
        # Arrange
        left = TestSpriteSheet(topleft=(0.49, 0.0))
        left.mask_seq = 4
        right = TestSpriteSheet(topleft=(2.0, 0.0))
        right.mask_seq = 4
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_mask_round_false_1(self):
        # Arrange
        left = TestSpriteSheet(topleft=(0.49, 0.0))
        left.mask_seq = 4
        right = TestSpriteSheet(topleft=(2.0, 0.0))
        right.mask_seq = 4
        # Act
        does_collide = right.does_collide(left)
        # Assert
        self.assertFalse(does_collide)

    def test_does_collide_mask_round_true_0(self):
        # Arrange
        left = TestSpriteSheet(topleft=(0.51, 0.0))
        left.mask_seq = 4
        right = TestSpriteSheet(topleft=(2.0, 0.0))
        right.mask_seq = 4
        # Act
        does_collide = left.does_collide(right)
        # Assert
        self.assertTrue(does_collide)

    def test_does_collide_mask_round_true_1(self):
        # Arrange
        left = TestSpriteSheet(topleft=(0.51, 0.0))
        left.mask_seq = 4
        right = TestSpriteSheet(topleft=(2.0, 0.0))
        right.mask_seq = 4
        # Act
        does_collide = right.does_collide(left)
        # Assert
        self.assertTrue(does_collide)

    def test_seq(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 6
        # Assert
        self.assertEqual(sprite.seq, 0)

    def test_seq_image_0(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 0
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('red'))

    def test_seq_image_1(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 1
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('green'))

    def test_seq_image_2(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 2
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('blue'))

    def test_seq_image_3(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 3
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('black'))

    def test_seq_image_4(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 4
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('white'))

    def test_seq_image_5(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.seq = 5
        # Assert
        self.assertEqual(sprite.image.size, (2, 2))
        for pos in self.IMAGE_POINTS:
            self.assertEqual(sprite.image.get_at(pos), pygame.Color('black'))

    def test_mask_seq(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 6
        # Assert
        self.assertEqual(sprite.mask_seq, 0)

    def test_mask_seq_mask_0(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 0
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 0)
        self.assertEqual(sprite.mask.get_at((1, 0)), 1)
        self.assertEqual(sprite.mask.get_at((0, 1)), 1)
        self.assertEqual(sprite.mask.get_at((1, 1)), 1)

    def test_mask_seq_mask_1(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 1
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 1)
        self.assertEqual(sprite.mask.get_at((1, 0)), 0)
        self.assertEqual(sprite.mask.get_at((0, 1)), 1)
        self.assertEqual(sprite.mask.get_at((1, 1)), 1)

    def test_mask_seq_mask_2(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 2
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 1)
        self.assertEqual(sprite.mask.get_at((1, 0)), 1)
        self.assertEqual(sprite.mask.get_at((0, 1)), 0)
        self.assertEqual(sprite.mask.get_at((1, 1)), 1)

    def test_mask_seq_mask_3(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 3
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 1)
        self.assertEqual(sprite.mask.get_at((1, 0)), 1)
        self.assertEqual(sprite.mask.get_at((0, 1)), 1)
        self.assertEqual(sprite.mask.get_at((1, 1)), 0)

    def test_mask_seq_mask_4(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 4
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 1)
        self.assertEqual(sprite.mask.get_at((1, 0)), 1)
        self.assertEqual(sprite.mask.get_at((0, 1)), 1)
        self.assertEqual(sprite.mask.get_at((1, 1)), 1)

    def test_mask_seq_mask_5(self):
        # Arrange
        sprite = TestSpriteSheet()
        # Act
        sprite.mask_seq = 5
        # Assert
        self.assertEqual(sprite.mask.get_size(), (2, 2))
        self.assertEqual(sprite.mask.get_at((0, 0)), 0)
        self.assertEqual(sprite.mask.get_at((1, 0)), 0)
        self.assertEqual(sprite.mask.get_at((0, 1)), 0)
        self.assertEqual(sprite.mask.get_at((1, 1)), 0)


if __name__ == '__main__':
    unittest.main()



import unittest
import os

import pygame

import jovialengine.gameinput as gameinput


class TestGameInput(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        gameinput.init(
            os.path.join('.', 'input.cfg'),
            1,
            (
                "Left",
                "Right",
                "Up",
                "Down",
                "Confirm",
                "Cancel",
            ),
            (
                gameinput.InputDefault(0, gameinput.EVENT_TYPE_START_POS + 0, gameinput.InputType.KEYBOARD, pygame.K_a),
                gameinput.InputDefault(0, gameinput.EVENT_TYPE_START_POS + 1, gameinput.InputType.KEYBOARD, pygame.K_d),
                gameinput.InputDefault(0, gameinput.EVENT_TYPE_START_POS + 2, gameinput.InputType.KEYBOARD, pygame.K_w),
                gameinput.InputDefault(0, gameinput.EVENT_TYPE_START_POS + 3, gameinput.InputType.KEYBOARD, pygame.K_s),
            )
        )

    def test__controller_states_init(self):
        # Assert
        self.assertEqual(len(gameinput._controller_states), 1)
        self.assertEqual(len(gameinput._controller_states[0]), 8)

    def test__controller_states_prev_init(self):
        # Assert
        self.assertEqual(len(gameinput._controller_states_prev), 1)
        self.assertEqual(len(gameinput._controller_states_prev[0]), 8)

    def test__input_mapping_init(self):
        # Arrange
        expected_input_mapping = {
            (gameinput.InputType.KEYBOARD, pygame.K_ESCAPE, 0): (0, gameinput.TYPE_PAUSE),
            (gameinput.InputType.KEYBOARD, pygame.K_F12, 0): (0, gameinput.TYPE_SCREENSHOT),
            (gameinput.InputType.CON_BUTTON, gameinput._CONTROLLER_PAUSE_BUTTON, 0): (0, gameinput.TYPE_PAUSE),
            (gameinput.InputType.KEYBOARD, pygame.K_a, 0): (0, gameinput.EVENT_TYPE_START_POS + 0),
            (gameinput.InputType.KEYBOARD, pygame.K_d, 0): (0, gameinput.EVENT_TYPE_START_POS + 1),
            (gameinput.InputType.KEYBOARD, pygame.K_w, 0): (0, gameinput.EVENT_TYPE_START_POS + 2),
            (gameinput.InputType.KEYBOARD, pygame.K_s, 0): (0, gameinput.EVENT_TYPE_START_POS + 3),
        }
        # Assert
        self.assertEqual(gameinput._input_mapping, expected_input_mapping)

    def test__get_input_id_display_keyboard(self):
        # Assert
        self.assertEqual(gameinput._get_input_id_display(gameinput.InputType.KEYBOARD, pygame.K_a), "a")

    def test__get_input_id_display_hat(self):
        # Assert
        self.assertEqual(gameinput._get_input_id_display(gameinput.InputType.CON_HAT, 2), "0U")

    def test_get_event_with_controls(self):
        # Assert
        self.assertEqual(gameinput.get_event_with_controls(0, gameinput.EVENT_TYPE_START_POS + 2), "Up: KEY-w")

    def test_get_event_name(self):
        # Assert
        self.assertEqual(gameinput.get_event_name(gameinput.EVENT_TYPE_START_POS + 3), "Down")

    def test_save_and_load(self):
        # Arrange
        expected = gameinput._input_mapping
        # Act
        gameinput.save()
        gameinput._load()
        # Assert
        self.assertEqual(gameinput._input_mapping, expected)

    def test_take_events(self):
        # Arrange
        events = [
            pygame.event.Event(pygame.KEYDOWN, {'key': pygame.K_d}),
            pygame.event.Event(pygame.KEYDOWN, {'key': pygame.K_s}),
        ]
        # Act
        gameinput.take_events(events)
        input_frame = gameinput.get_input_frame()
        # Assert
        self.assertEqual(input_frame._states, [[0, 0, 0, 1, 0, 1, 0, 0,]])
        self.assertTrue(input_frame.was_player_input_pressed(0, gameinput.EVENT_TYPE_START_POS + 1))
        self.assertTrue(input_frame.was_input_pressed(gameinput.EVENT_TYPE_START_POS + 3))


if __name__ == '__main__':
    unittest.main()



import pygame


class OffsetGroup(pygame.sprite.LayeredUpdates):
    """A sprite group for drawing sprites, in layers, offset by some amount."""
    def draw(self, surface: pygame.Surface, offset: pygame.typing.IntPoint=(0,0)):
        sprite_sequence = [
            (
                sprite.image,
                (
                    round(sprite.rect.x) + offset[0],
                    round(sprite.rect.y) + offset[1],
                )
            )
            for sprite
            in self.sprites()
        ]
        surface.fblits(sprite_sequence)
        self.lostsprites = []
        return self.lostsprites



from functools import cache

import pygame


@cache
def image(filename: str, alpha_or_colorkey: bool | pygame.typing.ColorLike=False):
    """Loads an image, converts, and sets colorkey as needed.
    The results are cached so don't alter them."""
    result = pygame.image.load(filename)
    if alpha_or_colorkey is True:
        result = result.convert_alpha()
    else:
        result = result.convert()
        if alpha_or_colorkey is not False:
            result.set_colorkey(alpha_or_colorkey)
    return result


@cache
def subsurface(surface: pygame.Surface, rect: tuple[int, int, int, int]):
    """Gets a subsurface from a surface.
    The results are cached so don't alter them."""
    return surface.subsurface(rect)


@cache
def flip(surface: pygame.Surface, flip_x: bool, flip_y: bool):
    """Gets a flipped surface from a surface.
    The results are cached so don't alter them."""
    return pygame.transform.flip(surface, flip_x, flip_y)


@cache
def mask_surface(surface: pygame.Surface, rect: tuple[int, int, int, int] | None=None):
    """Constructs a mask from a surface.
    The results are cached so don't alter them."""
    if rect:
        surface = subsurface(surface, rect)
    return pygame.mask.from_surface(surface)


@cache
def mask_filled(size: tuple[int, int]):
    """Constructs a filled mask.
    The results are cached so don't alter them."""
    return pygame.mask.Mask(size, True)


@cache
def mask_circle(size: tuple[int, int], radius: float):
    """Constructs a mask with a filled circle centered in the size at the given radius.
    The results are cached so don't alter them."""
    surface = pygame.Surface(size)
    surface.fill((0, 0, 0))
    pygame.draw.circle(surface, (255, 0, 0), (size[0] / 2, size[1] / 2), radius)
    surface.set_colorkey((0, 0, 0))
    return pygame.mask.from_surface(surface)


@cache
def sound(filename: str):
    """Loads a sound.
    The results are cached so don't alter them."""
    return pygame.mixer.Sound(filename)



import random
import math
import datetime

import pygame


def get_int_movement(tracking: float, vel: float, dt: int):
    tracking += vel * dt
    tracking_int = int(tracking)
    tracking -= tracking_int
    return tracking, tracking_int


def reduce_number(number, divisor):
    result = number // divisor
    mod = number % divisor
    if random.random() < (mod / divisor):
        result += 1
    return result


def sin_curve(number: float):
    return math.sin(number * math.pi / 2)


def cos_curve(number: float):
    if number == 1:
        return 0
    return math.cos(number * math.pi / 2)


def get_positional_channel_mix(pos: pygame.typing.Point, camera: pygame.FRect):
    lr_pos = pygame.math.clamp((pos[0] - camera.left) / camera.width, 0, 1)
    # currently doesn't start to get quieter as sprites get further off-screen to left or right (or up or down)
    # possibly should
    channel_l = _bound_channel_volume(cos_curve(lr_pos))
    channel_r = _bound_channel_volume(sin_curve(lr_pos))
    return channel_l, channel_r


def _bound_channel_volume(volume: float):
    return .2 + (volume * .8)


def binary(start, end, mix):
    if mix < 1.0:
        return start
    return end


def lerp(start, end, mix):
    return start + (end - start)*mix


def inc_speed_lerp(start, end, mix):
    return lerp(start, end, mix**2)


def dec_speed_lerp(start, end, mix):
    return lerp(start, end, 1 - (1 - mix)**2)


def inc_dec_speed_lerp(start, end, mix):
    midpoint = lerp(start, end, 0.5)
    if mix < 0.5:
        return inc_speed_lerp(start, midpoint, mix * 2)
    else:
        return dec_speed_lerp(midpoint, end, (mix - 0.5) * 2)


def dec_inc_speed_lerp(start, end, mix):
    midpoint = lerp(start, end, 0.5)
    if mix < 0.5:
        return dec_speed_lerp(start, midpoint, mix * 2)
    else:
        return inc_speed_lerp(midpoint, end, (mix - 0.5) * 2)


def get_datetime_file_name():
    return datetime.datetime.now(datetime.UTC).isoformat().replace(':', '')



import abc
from typing import final, TYPE_CHECKING
from collections.abc import Iterable

import pygame

from . import display
from .offsetgroup import OffsetGroup
from .inputframe import InputFrame
if TYPE_CHECKING:
    from .gamesprite import GameSprite


class ModeBase(abc.ABC):
    """Base class for all game modes.
    Subclasses should set:
    optional: _SPACE_SIZE, size of the space inside this mode, if not supplied will assume display.screen_size
    optional: _CAMERA_SIZE, size of the camera inside this mode, if not supplied will assume display.screen_size
    optional: _CAMERA_OFFSET, offset for drawing camera view onto screen

    When a subclass wants to pass on to another mode, set self.next_mode.
    Don't create another mode unless you are immediately assigning it to self.next_mode.
    """
    _SPACE_SIZE: tuple[int, int] | None = None
    _CAMERA_SIZE: tuple[int, int] | None = None
    _CAMERA_OFFSET: tuple[int, int] = (0, 0)

    __slots__ = (
        '_background',
        'sprites_all',
        'sprites_game',
        'sprites_input',
        '_camera',
        '_input_frame',
        'next_mode',
    )

    def __init__(self):
        self._background = pygame.Surface(self.get_space_size()).convert()
        self._background.fill((0, 0, 0))
        self.sprites_all = OffsetGroup()
        self.sprites_game: pygame.sprite.Group[GameSprite] = pygame.sprite.Group()
        self.sprites_input: pygame.sprite.Group[GameSprite] = pygame.sprite.Group()
        self._camera = pygame.FRect((0, 0), self._CAMERA_SIZE or display.screen_size)
        self._input_frame: InputFrame | None = None
        self.next_mode: ModeBase | None = None

    @final
    def input(self, events: Iterable[pygame.event.Event], input_frame: InputFrame):
        """All game modes can take in input."""
        for event in events:
            self._take_event(event)
        self._take_frame(input_frame)
        for sprite in self.sprites_input.sprites():
            sprite.input(input_frame)
        self._input_frame = input_frame

    @final
    def update(self, dt: int):
        """All game modes can update."""
        self._update_pre_sprites(dt)
        for sprite in self.sprites_all.sprites():
            sprite.update(dt, self._camera)
        self._update_post_sprites(dt)
        self.__handle_collisions()

    def __handle_collisions(self):
        collide_events = []
        collide_sprites = self.sprites_game.sprites()
        for i, sprite0 in enumerate(collide_sprites):
            for j in range(i + 1, len(collide_sprites)):
                sprite1 = collide_sprites[j]
                sprite0_collides = sprite0.get_collides_with() & sprite1.get_collision_labels()
                sprite1_collides = sprite1.get_collides_with() & sprite0.get_collision_labels()
                if (sprite0_collides or sprite1_collides) and sprite0.does_collide(sprite1):
                    for sprite0_collide in sprite0_collides:
                        collide_events.append((getattr(sprite0, 'collide_' + sprite0_collide), sprite1,))
                    for sprite1_collide in sprite1_collides:
                        collide_events.append((getattr(sprite1, 'collide_' + sprite1_collide), sprite0,))
        for collide_event in collide_events:
            collide_event[0](collide_event[1])

    @final
    def draw(self, screen: pygame.Surface):
        """All game modes can draw to the screen"""
        self._update_pre_draw()
        screen.set_clip((self._CAMERA_OFFSET, self._CAMERA_SIZE or display.screen_size))
        offset = (
            self._CAMERA_OFFSET[0] - round(self._camera.x),
            self._CAMERA_OFFSET[1] - round(self._camera.y),
        )
        screen.blit(self._background, offset)
        self._draw_pre_sprites(screen, offset)
        self.sprites_all.draw(screen, offset)
        for sprite in self.sprites_game.sprites():
            sprite.draw_dynamic(screen, offset)
        self._draw_post_sprites(screen, offset)
        screen.set_clip(None)
        self._draw_post_camera(screen)

    @final
    def cleanup(self):
        for sprites in (self.sprites_all, self.sprites_game, self.sprites_input):
            # we can't just kill the sprites since we might be reusing them between modes
            sprites.empty()
        self._cleanup()

    def _take_event(self, event: pygame.event.Event):
        """Handle any input that requires looking at pygame events directly, like typing."""
        pass

    def _take_frame(self, input_frame: InputFrame):
        """Handle all other input.
        During this method call self._input_frame still holds the old input_frame."""
        pass

    def _update_pre_sprites(self, dt: int):
        """Handle mode updates before sprites."""
        pass

    def _update_post_sprites(self, dt: int):
        """Handle mode updates after sprites (before collision handling)."""
        pass

    def _update_pre_draw(self):
        """Handle updates that only need to happen right before drawing, like updating the camera position."""
        pass

    def _draw_pre_sprites(self, screen: pygame.Surface, offset: pygame.typing.IntPoint):
        """Handle dynamic drawing after the background and before sprites."""
        pass

    def _draw_post_sprites(self, screen: pygame.Surface, offset: pygame.typing.IntPoint):
        """Handle dynamic drawing after the sprites."""
        pass

    def _draw_post_camera(self, screen: pygame.Surface):
        """Handle drawing onto screen after camera-aware drawing is done."""
        pass

    def _cleanup(self):
        """Handle any additional cleanup this mode will need when it's ended."""
        pass

    @staticmethod
    def _stop_mixer():
        pygame.mixer.music.stop()
        pygame.mixer.music.unload()
        pygame.mixer.stop()

    def get_space_size(self):
        return self._SPACE_SIZE or display.screen_size



import unittest

import pygame

import jovialengine.utility as utility


class TestUtility(unittest.TestCase):
    def test_get_int_movement(self):
        # Arrange
        tracking = 0.6
        # Act
        tracking, tracking_int = utility.get_int_movement(tracking, 3.75, 10)
        # Assert
        self.assertEqual(tracking_int, 38)
        self.assertTrue(0.0999999999 < tracking < 0.1000000001)

    def test_get_positional_channel_mix(self):
        # Arrange
        pos = (100, 100)
        camera = pygame.Rect(10, 0, 320, 240)
        # Act
        result = utility.get_positional_channel_mix(pos, camera)
        # Assert
        self.assertEqual(result, (0.9231914344987546, 0.5420440747442257))

    def test_binary(self):
        # Assert
        self.assertEqual(
            utility.binary(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 0.99999),
            1
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 1.0),
            2
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 1),
            2
        )

    def test_lerp(self):
        # Assert
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.25),
            1.25
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.75),
            1.75
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 1.0),
            2
        )

    def test_inc_speed_lerp(self):
        # Assert
        self.assertEqual(
            utility.inc_speed_lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.inc_speed_lerp(1.0, 2.0, 0.25),
            1.0625
        )
        self.assertEqual(
            utility.inc_speed_lerp(1.0, 2.0, 0.75),
            1.5625
        )
        self.assertEqual(
            utility.inc_speed_lerp(1.0, 2.0, 1.0),
            2
        )

    def test_dec_speed_lerp(self):
        # Assert
        self.assertEqual(
            utility.dec_speed_lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.dec_speed_lerp(1.0, 2.0, 0.25),
            1.4375
        )
        self.assertEqual(
            utility.dec_speed_lerp(1.0, 2.0, 0.75),
            1.9375
        )
        self.assertEqual(
            utility.dec_speed_lerp(1.0, 2.0, 1.0),
            2
        )

    def test_inc_dec_speed_lerp(self):
        # Assert
        self.assertEqual(
            utility.inc_dec_speed_lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.inc_dec_speed_lerp(1.0, 2.0, 0.25),
            1.125
        )
        self.assertEqual(
            utility.inc_dec_speed_lerp(1.0, 2.0, 0.75),
            1.875
        )
        self.assertEqual(
            utility.inc_dec_speed_lerp(1.0, 2.0, 1.0),
            2
        )

    def test_dec_inc_speed_lerp(self):
        # Assert
        self.assertEqual(
            utility.dec_inc_speed_lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.dec_inc_speed_lerp(1.0, 2.0, 0.25),
            1.375
        )
        self.assertEqual(
            utility.dec_inc_speed_lerp(1.0, 2.0, 0.75),
            1.625
        )
        self.assertEqual(
            utility.dec_inc_speed_lerp(1.0, 2.0, 1.0),
            2
        )


if __name__ == '__main__':
    unittest.main()



import unittest

import pygame

import jovialengine.display as display


class TestDisplay(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        display.upscale = 3
        display.screen_size = (320, 240)

    @staticmethod
    def set_up_full_screen():
        display.is_fullscreen = True
        display._fullscreen_offset = [50, 100]

    @staticmethod
    def set_up_windowed():
        display.is_fullscreen = False
        display._fullscreen_offset = None

    def test_scale_mouse_input_fullscreen_button(self):
        # Arrange
        self.set_up_full_screen()
        event_dict = {
            'pos': (400, 800),
            'button': 1,
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result_event = display.scale_mouse_input(event)
        # Assert
        self.assertEqual(result_event.pos, (116, 233))
        self.assertEqual(result_event.button, 1)

    def test_scale_mouse_input_fullscreen_motion(self):
        # Arrange
        self.set_up_full_screen()
        event_dict = {
            'pos': (400, 800),
            'rel': (100, 200),
            'buttons': (1,),
        }
        event = pygame.event.Event(pygame.MOUSEMOTION, event_dict)
        # Act
        result_event = display.scale_mouse_input(event)
        # Assert
        self.assertEqual(result_event.pos, (116, 233))
        self.assertEqual(result_event.rel, (33, 66))
        self.assertEqual(result_event.buttons, (1,))

    def test_scale_mouse_input_windowed_button(self):
        # Arrange
        self.set_up_windowed()
        event_dict = {
            'pos': (400, 800),
            'button': 1,
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result_event = display.scale_mouse_input(event)
        # Assert
        self.assertEqual(result_event.pos, (133, 266))
        self.assertEqual(result_event.button, 1)

    def test_scale_mouse_input_windowed_motion(self):
        # Arrange
        self.set_up_windowed()
        event_dict = {
            'pos': (400, 800),
            'rel': (100, 200),
            'buttons': (1,),
        }
        event = pygame.event.Event(pygame.MOUSEMOTION, event_dict)
        # Act
        result_event = display.scale_mouse_input(event)
        # Assert
        self.assertEqual(result_event.pos, (133, 266))
        self.assertEqual(result_event.rel, (33, 66))
        self.assertEqual(result_event.buttons, (1,))

    def test_is_in_screen_true(self):
        # Arrange
        self.set_up_windowed()
        # Act
        result = display.is_in_screen((2, 3))
        # Assert
        self.assertTrue(result)

    def test_is_in_screen_false(self):
        # Arrange
        self.set_up_windowed()
        # Act
        result = display.is_in_screen((-2, 3))
        # Assert
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()
